@@ -25,7 +25,7 @@
 #include <string.h>
 #include <strings.h>
 #include <popt.h>
-#include <gdk/gdkwindow.h>
+#include <gdk/gdk.h>
 #include <libbonobo.h>
 #include <login-helper/login-helper.h>
 #include <gdk/gdkx.h>
@@ -40,10 +40,12 @@
 #include "zoom-region-private.h"
 #include "GNOME_Magnifier.h"
 
-#undef DEBUG_CLIENT_CALLS
+/* if you #define this, don't forget to set MAG_CLIENT_DEBUG env variable */
+#define DEBUG_CLIENT_CALLS
 
 #ifdef DEBUG_CLIENT_CALLS
-#define DBG(a) a
+static gboolean client_debug = FALSE;
+#define DBG(a) if (client_debug) { (a); }
 #else
 #define DBG(a)
 #endif
@@ -92,6 +94,24 @@ enum {
 	MAGNIFIER_CROSSWIRE_COLOR_PROP
 } PropIdx;
 
+typedef struct
+{
+	GNOME_Magnifier_RectBounds rectbounds;
+	GNOME_Magnifier_RectBounds viewport;
+	gboolean is_managed;
+	gint scroll_policy;
+	gfloat contrast;
+	gfloat zx;
+	gfloat zy;
+	gint32 xalign;
+	gint32 yalign;
+	guint32 border_color;
+	gint32 border_size;
+	gchar *smoothing_type;
+	gboolean inverse;
+
+} MagnifierZoomRegionSaveProps;
+
 #ifdef DEBUG_CLIENT_CALLS
 gchar* mag_prop_names[MAGNIFIER_CROSSWIRE_COLOR_PROP + 1] = {
     "SOURCE_DISPLAY",
@@ -113,12 +133,15 @@ gchar* mag_prop_names[MAGNIFIER_CROSSWIRE_COLOR_PROP + 1] = {
 static int _x_error = 0;
 static int fixes_event_base = 0, fixes_error_base;
 static Display *cursor_client_connection;
+static guint    cursor_client_gsource = 0;
 static Magnifier *_this_magnifier = NULL;
 
 static void magnifier_transform_cursor (Magnifier *magnifier);
 static void magnifier_init_cursor_set (Magnifier *magnifier, gchar *cursor_set);
+static void magnifier_init_window (Magnifier *magnifier, GdkScreen *screen);
 static gboolean magnifier_check_set_struts (Magnifier *magnifier);
 static gboolean magnifier_reset_struts_at_idle (gpointer data);
+static void magnifier_init_window (Magnifier *magnifier, GdkScreen *screen);
 static gboolean _is_override_redirect = FALSE;
 
 static Window*
@@ -248,6 +271,7 @@ magnifier_zoom_regions_mark_dirty (Magnifier *magnifier, GNOME_Magnifier_RectBou
 	GList *list;
 
 	g_assert (magnifier);
+
 	list = magnifier->zoom_regions;
 	while (list) 
 	{
@@ -318,6 +342,19 @@ magnifier_free_cursor_pixels (guchar *pixels, gpointer data)
     /* XFree (data); FIXME why doesn't this work properly? */
 }
 
+#ifdef HAVE_XFIXES
+static void
+magnifier_cursor_convert_to_rgba (Magnifier *magnifier, XFixesCursorImage *cursor_image)
+{
+	int i, count = cursor_image->width * cursor_image->height;
+	for (i = 0; i < count; ++i) 
+	{
+		guint32 pixval = GUINT_TO_LE (cursor_image->pixels[i]);
+		cursor_image->pixels[i] = pixval;
+	}
+}
+#endif
+
 GdkPixbuf *
 magnifier_get_source_pixbuf (Magnifier *magnifier)
 {
@@ -327,6 +364,7 @@ magnifier_get_source_pixbuf (Magnifier *magnifier)
 	gchar s[6];
 	if (cursor_image)
 	{
+	        magnifier_cursor_convert_to_rgba (magnifier, cursor_image);
 		cursor_pixbuf = gdk_pixbuf_new_from_data ((guchar *) cursor_image->pixels, 
 							  GDK_COLORSPACE_RGB, 
 							  TRUE, 8,  
@@ -434,8 +472,15 @@ magnifier_cursor_notification_init (Magnifier *magnifier)
 
     if (!magnifier->source_display) return FALSE;
 
-    dpy = GDK_DISPLAY_XDISPLAY (magnifier->source_display);
-    cursor_client_connection = XOpenDisplay (0);
+    if (cursor_client_connection)
+    {
+        /* remove the old watch */
+	if (cursor_client_gsource) 
+	  g_source_remove (cursor_client_gsource);
+	XCloseDisplay (cursor_client_connection);
+    }
+
+    cursor_client_connection = XOpenDisplay (magnifier->source_display_name);
     rootwin = GDK_WINDOW_XWINDOW (magnifier->priv->root);
 
     if (!XFixesQueryExtension (cursor_client_connection, &fixes_event_base, &fixes_error_base))
@@ -448,8 +493,9 @@ magnifier_cursor_notification_init (Magnifier *magnifier)
 	XFixesSelectCursorInput (cursor_client_connection, rootwin, XFixesDisplayCursorNotifyMask);
 	fd = ConnectionNumber (cursor_client_connection);
 	ioc = g_io_channel_unix_new (fd);
-	g_io_add_watch (ioc, G_IO_IN | G_IO_HUP | G_IO_PRI | G_IO_ERR, magnifier_cursor_notify, 
-			magnifier);
+	cursor_client_gsource = 
+	    g_io_add_watch (ioc, G_IO_IN | G_IO_HUP | G_IO_PRI | G_IO_ERR, magnifier_cursor_notify, 
+			    magnifier);
 	g_io_channel_unref (ioc); 
 	g_message ("added event source to xfixes cursor-notify connection");
  	XFlush (cursor_client_connection); 
@@ -469,6 +515,7 @@ magnifier_notify_damage (Magnifier *magnifier, XRectangle *rect)
 	rect_bounds.y1 = rect->y;
 	rect_bounds.x2 = rect->x + rect->width;
 	rect_bounds.y2 = rect->y + rect->height;
+#undef DEBUG_DAMAGE
 #ifdef DEBUG_DAMAGE
 	g_message ("damage");
 	g_message ("dirty %d, %d to %d, %d", rect_bounds.x1, rect_bounds.y1, rect_bounds.x2, rect_bounds.y2);
@@ -631,16 +678,25 @@ magnifier_unref_zoom_region (gpointer data, gpointer user_data)
 	CORBA_Environment ev;
 	GNOME_Magnifier_ZoomRegion zoom_region = data;
 	CORBA_exception_init (&ev);
+	
+	DBG(g_message ("unreffing zoom region"));
+
 	GNOME_Magnifier_ZoomRegion_dispose (zoom_region, &ev);
+	if (!BONOBO_EX (&ev))
+	    Bonobo_Unknown_unref (zoom_region, &ev);
 }
 
-static void
-magnifier_reparent_zoom_regions (Magnifier *magnifier)
+static GSList*
+magnifier_zoom_regions_save (Magnifier *magnifier)
 {
     GList *list;
+    GSList *save_props = NULL;
     
     g_assert (magnifier);
     list = magnifier->zoom_regions;
+
+    DBG(g_message ("saving %d regions", g_list_length (list)));
+
     while (list) 
     {
 	GNOME_Magnifier_ZoomRegion zoom_region;
@@ -649,47 +705,81 @@ magnifier_reparent_zoom_regions (Magnifier *magnifier)
 	CORBA_exception_init (&ev);
 	if (zoom_region)
 	{
-	    GNOME_Magnifier_ZoomRegion new_region;
-	    Bonobo_PropertyBag properties, new_properties;
-	    GNOME_Magnifier_RectBounds rectbounds, viewport;
+	    Bonobo_PropertyBag properties;
 	    CORBA_any *value;
-	    rectbounds = GNOME_Magnifier_ZoomRegion_getROI (zoom_region, &ev);
+	    MagnifierZoomRegionSaveProps *zoomer_props = g_new0 (MagnifierZoomRegionSaveProps, 1);
+
+	    zoomer_props->rectbounds = GNOME_Magnifier_ZoomRegion_getROI (zoom_region, &ev);
 	    properties = GNOME_Magnifier_ZoomRegion_getProperties (zoom_region, &ev);
 	    value = bonobo_pbclient_get_value (properties, "viewport", TC_GNOME_Magnifier_RectBounds, &ev);
-	    memcpy (&viewport, value->_value, sizeof (GNOME_Magnifier_RectBounds));
-	    fprintf (stderr, "reparenting zoomer with viewport %d,%d - %d,%d\n", viewport.x1, viewport.y1, viewport.x2, viewport.y2);
+	    memcpy (&zoomer_props->viewport, value->_value, sizeof (GNOME_Magnifier_RectBounds));
 	    CORBA_free (value);
-	    new_region = GNOME_Magnifier_Magnifier_createZoomRegion (BONOBO_OBJREF (magnifier), 0, 0, &rectbounds, &viewport, &ev);
-	    new_properties = GNOME_Magnifier_ZoomRegion_getProperties (new_region, &ev);
-	    bonobo_pbclient_set_boolean (new_properties, "is-managed", 
-					 bonobo_pbclient_get_boolean (properties, "is-managed", NULL), NULL);
-	    bonobo_pbclient_set_long (new_properties, "smooth-scroll-policy", 
-				       bonobo_pbclient_get_long (properties, "smooth-scroll-policy", NULL), NULL);
-	    bonobo_pbclient_set_float (new_properties, "contrast", 
-					 bonobo_pbclient_get_float (properties, "contrast", NULL), NULL);
-	    bonobo_pbclient_set_float (new_properties, "mag-factor-x", 
-					 bonobo_pbclient_get_float (properties, "mag-factor-x", NULL), NULL);
-	    bonobo_pbclient_set_float (new_properties, "mag-factor-y", 
-					 bonobo_pbclient_get_float (properties, "mag-factor-y", NULL), NULL);
-	    bonobo_pbclient_set_long (new_properties, "x-alignment", 
-					 bonobo_pbclient_get_long (properties, "x-alignment", NULL), NULL);
-	    bonobo_pbclient_set_long (new_properties, "y-alignment", 
-					 bonobo_pbclient_get_long (properties, "y-alignment", NULL), NULL);
-	    bonobo_pbclient_set_long (new_properties, "border-color", 
-				      bonobo_pbclient_get_long (properties, "border-color", NULL), NULL); 
-	    bonobo_pbclient_set_long (new_properties, "border-size", 
-				      bonobo_pbclient_get_long (properties, "border-size", NULL), NULL);
-	    bonobo_pbclient_set_string (new_properties, "smoothing-type", 
-					bonobo_pbclient_get_string (properties, "smoothing-type", NULL), NULL); 
-	    bonobo_pbclient_set_boolean (new_properties, "inverse-video", 
-					 bonobo_pbclient_get_boolean (properties, "inverse-video", NULL), NULL); 
+	    zoomer_props->is_managed = bonobo_pbclient_get_boolean (properties, "is-managed", NULL);
+	    zoomer_props->scroll_policy = bonobo_pbclient_get_short (properties, "smooth-scroll-policy", NULL);
+	    zoomer_props->contrast = bonobo_pbclient_get_float (properties, "contrast", NULL);
+	    zoomer_props->zx = bonobo_pbclient_get_float (properties, "mag-factor-x", NULL);
+	    zoomer_props->zy = bonobo_pbclient_get_float (properties, "mag-factor-y", NULL);
+	    zoomer_props->xalign = bonobo_pbclient_get_long (properties, "x-alignment", NULL);
+	    zoomer_props->yalign = bonobo_pbclient_get_long (properties, "y-alignment", NULL);
+	    zoomer_props->border_color = bonobo_pbclient_get_long (properties, "border-color", NULL); 
+	    zoomer_props->border_size = bonobo_pbclient_get_long (properties, "border-size", NULL);
+	    zoomer_props->smoothing_type = bonobo_pbclient_get_string (properties, "smoothing-type", NULL); 
+	    zoomer_props->inverse = bonobo_pbclient_get_boolean (properties, "inverse-video", NULL); 
+
 	    bonobo_object_release_unref (properties, &ev);
-	    bonobo_object_release_unref (new_properties, &ev);
-	    list->data = new_region;
 	    magnifier_unref_zoom_region ((gpointer) zoom_region, NULL);
+	    save_props = g_slist_append (save_props, zoomer_props);
 	}
 	list = g_list_next (list);
     }   
+
+    magnifier->zoom_regions = NULL;
+
+    return save_props;
+}
+
+static void
+magnifier_zoom_regions_restore (Magnifier *magnifier, GSList *region_params)
+{
+	GSList *list = region_params;
+
+	while (list)
+	{
+		CORBA_Environment ev;
+		MagnifierZoomRegionSaveProps *zoomer_props = list->data;
+		GNOME_Magnifier_ZoomRegion new_region;
+		Bonobo_PropertyBag new_properties;
+
+		CORBA_exception_init (&ev);
+		new_region = GNOME_Magnifier_Magnifier_createZoomRegion (BONOBO_OBJREF (magnifier), zoomer_props->zx, zoomer_props->zy, &zoomer_props->rectbounds, &zoomer_props->viewport, &ev);
+		new_properties = GNOME_Magnifier_ZoomRegion_getProperties (new_region, &ev);
+		bonobo_pbclient_set_boolean (new_properties, "is-managed", 
+					     zoomer_props->is_managed, NULL);
+		bonobo_pbclient_set_short (new_properties, "smooth-scroll-policy", 
+					   zoomer_props->scroll_policy, NULL);
+		bonobo_pbclient_set_float (new_properties, "contrast", 
+					   zoomer_props->contrast, NULL);
+/* NOT YET USED
+		bonobo_pbclient_set_long (new_properties, "x-alignment", 
+					     zoomer_props->xalign, NULL);
+		bonobo_pbclient_set_long (new_properties, "y-alignment", 
+					     zoomer_props->yalign, NULL);
+*/
+		bonobo_pbclient_set_long (new_properties, "border-color", 
+					     zoomer_props->border_color, NULL);
+		bonobo_pbclient_set_long (new_properties, "border-size", 
+					     zoomer_props->border_size, NULL);
+		bonobo_pbclient_set_string (new_properties, "smoothing-type", 
+					     zoomer_props->smoothing_type, NULL);
+		bonobo_pbclient_set_boolean (new_properties, "inverse-video", 
+					     zoomer_props->inverse, NULL);
+		GNOME_Magnifier_Magnifier_addZoomRegion (BONOBO_OBJREF (magnifier), new_region, &ev);
+		g_free (zoomer_props->smoothing_type);
+		g_free (zoomer_props);
+		bonobo_object_release_unref (new_properties, &ev);
+		list = g_slist_next (list);
+	}
+	g_slist_free (region_params);
 }
 
 static void
@@ -1124,6 +1214,7 @@ magnifier_set_property (BonoboPropertyBag *bag,
 		full_display_string = BONOBO_ARG_GET_STRING (arg);
 		if (can_open_display (full_display_string))
 		{
+		    GSList *zoom_region_params = NULL;
 		    magnifier->source_screen_num =
 			magnifier_parse_display_name (magnifier,
 						      full_display_string,
@@ -1131,16 +1222,20 @@ magnifier_set_property (BonoboPropertyBag *bag,
 		    magnifier->source_display =
 			gdk_display_open (full_display_string);
 		    magnifier->source_display_name = g_strdup (full_display_string);
+		    zoom_region_params = magnifier_zoom_regions_save (magnifier);
 		    magnifier->priv->root =
 			gdk_screen_get_root_window (
 			    gdk_display_get_screen (
 				magnifier->source_display,
 				magnifier->source_screen_num));
-		    magnifier_warp_cursor_to_screen (magnifier);
+	             /* attach listeners for DAMAGE, "dirty region", XFIXES cursor changes */
+	            magnifier_set_extension_listeners (magnifier, magnifier_get_root (magnifier));
 		    magnifier_get_display_rect_bounds (magnifier, &magnifier->source_bounds, FALSE);
+		    magnifier_zoom_regions_restore (magnifier, zoom_region_params);
+		    magnifier_warp_cursor_to_screen (magnifier);
 		    magnifier_check_set_struts (magnifier);
 		}
-		DBG (fprintf (stderr, "Set source display: \t%s\n", full_display_string));
+		DBG(fprintf (stderr, "Set source display: \t%s\n", full_display_string));
 		break;
 	case MAGNIFIER_TARGET_DISPLAY_PROP:
 		full_display_string = BONOBO_ARG_GET_STRING (arg);
@@ -1161,27 +1256,46 @@ magnifier_set_property (BonoboPropertyBag *bag,
 						   magnifier->target_display,
 						   magnifier->target_screen_num));
 #else
+			GSList *zoom_region_params = NULL;
 			/* disconnect from the old window's destroy signal */
 			g_object_disconnect (magnifier->priv->w,
 				  "any_signal::realize", magnifier_realize, NULL,
 				  "any_signal::size_allocate", magnifier_size_allocate, NULL,
 				  "any_signal::destroy", magnifier_exit, NULL,
 			          NULL);
-			/* destroy the old windows */
+			/* save the old zoom region state */
+		        zoom_region_params = magnifier_zoom_regions_save (magnifier);
+			/* destroy the old window */
 			gtk_widget_destroy (magnifier->priv->w);
 			/* and re-initialize... */
 			magnifier_init_window (magnifier, gdk_display_get_screen (
 						   magnifier->target_display,
 						   magnifier->target_screen_num));
+			/* restore the zoom regions in their new host magnifier window */
+			magnifier_zoom_regions_restore (magnifier, zoom_region_params);
 #endif
 		    }
-		    magnifier_get_display_rect_bounds (magnifier, &magnifier->target_bounds, TRUE);
 		    magnifier_get_display_rect_bounds (magnifier, &magnifier->source_bounds, FALSE);
-		    magnifier_reparent_zoom_regions (magnifier);
 		    magnifier_init_cursor_set (magnifier, magnifier->cursor_set); /* needed to reset pixmaps */
+		    gtk_window_move (GTK_WINDOW (magnifier->priv->w),
+				     magnifier->target_bounds.x1,
+				     magnifier->target_bounds.y1);
+		    
+		    if ((magnifier->target_bounds.x2 - magnifier->target_bounds.x1 > 0) &&
+			(magnifier->target_bounds.y2 - magnifier->target_bounds.y1) > 0)
+		    {
+			gtk_window_resize (GTK_WINDOW (magnifier->priv->w),
+				       magnifier->target_bounds.x2 - magnifier->target_bounds.x1,
+				       magnifier->target_bounds.y2 - magnifier->target_bounds.y1);
+		    DBG(fprintf (stderr, "Set target size: \t%d,%d to %d,%d\n", 
+			     magnifier->target_bounds.x1, magnifier->target_bounds.y1, magnifier->target_bounds.x2, magnifier->target_bounds.y2));
+		    }
+		    /* N. B. we don't reset the target bounds to the limits of the new display, because */
+		    /* doing so would override the client-specified magnifier size */
+		    /* magnifier_get_display_rect_bounds (magnifier, &magnifier->target_bounds, TRUE); */
 		    magnifier_check_set_struts (magnifier);
 		}
-		DBG (fprintf (stderr, "Set target display: \t%s (screen %d)\n", 
+		DBG(fprintf (stderr, "Set target display: \t%s (screen %d)\n", 
 			      full_display_string, magnifier->target_screen_num));
 		break;
 	case MAGNIFIER_SOURCE_SIZE_PROP:
@@ -1205,7 +1319,7 @@ magnifier_set_property (BonoboPropertyBag *bag,
 				   magnifier->target_bounds.x2 - magnifier->target_bounds.x1,
 				   magnifier->target_bounds.y2 - magnifier->target_bounds.y1);
 		magnifier_check_set_struts (magnifier);
-		DBG (fprintf (stderr, "Set target size: \t%d,%d to %d,%d\n", 
+		DBG(fprintf (stderr, "Set target size: \t%d,%d to %d,%d\n", 
 			      magnifier->target_bounds.x1, magnifier->target_bounds.y1, magnifier->target_bounds.x2, magnifier->target_bounds.y2));
 		break;
 	case MAGNIFIER_CURSOR_SET_PROP:
@@ -1301,8 +1415,6 @@ impl_magnifier_set_source_display (PortableServer_Servant servant,
 				    MAGNIFIER_SOURCE_DISPLAY_PROP,
 				    ev,
 				    magnifier);
-	    /* attach listeners for DAMAGE, "dirty region", XFIXES cursor changes */
-	    magnifier_set_extension_listeners (magnifier, magnifier_get_root (magnifier));
 	}
 	else
 	{
@@ -1323,7 +1435,7 @@ impl_magnifier_set_target_display (PortableServer_Servant servant,
 	
 	DBG (fprintf (stderr, "Set target display: \t%s\n", display));
 
-	if (strcmp (display, magnifier->source_display_name)) 
+	if (strcmp (display, magnifier->target_display_name)) 
 	{
 	    magnifier_set_property (magnifier->property_bag,
 				    arg,
@@ -1333,7 +1445,7 @@ impl_magnifier_set_target_display (PortableServer_Servant servant,
 	}
 	else
 	{
-	    DBG (fprintf (stderr, "Attempt to set source to same value as previous: %s\n",
+	    DBG (fprintf (stderr, "Attempt to set target to same value as previous: %s\n",
 			  display));
 	}
 	bonobo_arg_release (arg);
@@ -1357,9 +1469,9 @@ impl_magnifier_get_target_display (PortableServer_Servant servant,
 {
         Magnifier *magnifier = MAGNIFIER (bonobo_object_from_servant (servant));
 	DBG (fprintf (stderr, "Get target display: \t%s\n", 
-		      magnifier->source_display_name));
+		      magnifier->target_display_name));
 
-	return CORBA_string_dup (magnifier->source_display_name ? magnifier->target_display_name : "");
+	return CORBA_string_dup (magnifier->target_display_name ? magnifier->target_display_name : "");
 }
 
 static GNOME_Magnifier_ZoomRegion
@@ -1382,6 +1494,7 @@ impl_magnifier_create_zoom_region (PortableServer_Servant servant,
 	 * shouldn't do this here, since it causes the region to get
 	 * mapped onto the parent, if if it's not explicitly added!
 	 */
+	DBG(g_message ("creating zoom region with parent %p", magnifier));
 	zoom_region->priv->parent = magnifier;
 
 	retval = BONOBO_OBJREF (zoom_region);
@@ -1481,7 +1594,7 @@ impl_magnifier_clear_all_zoom_regions (PortableServer_Servant servant,
 				       CORBA_Environment * ev)
 {
 	Magnifier *magnifier = MAGNIFIER (bonobo_object_from_servant (servant));
-	DBG (fprintf (stderr, "Clear all zoom regions.\n"));
+	fprintf (stderr, "Clear all zoom regions.\n");
 
 	g_list_foreach (magnifier->zoom_regions,
 			magnifier_unref_zoom_region, magnifier);
@@ -1502,6 +1615,7 @@ magnifier_class_init (MagnifierClass *klass)
 {
         GObjectClass * object_class = (GObjectClass *) klass;
         POA_GNOME_Magnifier_Magnifier__epv *epv = &klass->epv;
+	parent_class = g_type_class_peek (BONOBO_TYPE_OBJECT); /* needed by BONOBO_CALL_PARENT! */
 
 	object_class->dispose = magnifier_gobject_dispose;
 
@@ -1693,7 +1807,7 @@ magnifier_init_window (Magnifier *magnifier, GdkScreen *screen)
 				  "signal::size_allocate", magnifier_size_allocate, NULL,
 				  "signal::destroy", magnifier_exit, NULL,
 				  NULL);
-	gtk_window_set_screen (magnifier->priv->w, screen);
+	gtk_window_set_screen (GTK_WINDOW (magnifier->priv->w), screen);
 	magnifier->priv->canvas = gtk_fixed_new ();
 	gtk_container_add (GTK_CONTAINER (magnifier->priv->w),
 			   magnifier->priv->canvas);
@@ -1727,6 +1841,9 @@ magnifier_init (Magnifier *magnifier)
 			       gdk_display_get_screen (magnifier->target_display, 
 						       magnifier->target_screen_num));
 	magnifier_init_cursor_set (magnifier, "default");
+#ifdef DEBUG_CLIENT_CALLS
+	client_debug = (g_getenv ("MAG_CLIENT_DEBUG") != NULL);
+#endif
 }
 
 GdkDrawable *
